#include <thread>
#include <chrono>
#include <map>
#include <iostream>

#include <boost/algorithm/string.hpp>
#include <boost/lockfree/fifo.hpp>

#include <tbb/concurrent_queue.h>


using namespace std;
using namespace boost;

using namespace boost::lockfree;
using namespace tbb::strict_ppl;


void SimpleTest();

// Use Intel's Thread Building Blocks concurrent_queue<T>
// That is supposed to be non blocking too.
void SimpleTestTBB();


void TestManyThreads(int nConsumers);



int main(int argc, char** argv)
{
    fifo<int> test;

    test.enqueue(3);
    cout << "Is lock free ? " << test.is_lock_free() << endl;
    
    
    SimpleTest();
    // SimpleTestTBB();
    // TestManyThreads(4);
    
    return 0;
}



void TestManyThreads(int nConsumers)
{
    int count = 1 * 1000 * 1000;
    int nThreads = nConsumers;
    
    bool over = false;
    int totals[nThreads];

    map<int, int> results[nThreads];
        
    thread producers[nThreads];
    thread consumers[nThreads];
    
    fifo<int> test;
    
    int produced = 0;
    int consumed = 0;
    
    // Producers
    for (int i = 0; i < nThreads; ++i) {
        producers[i] = thread([&] () -> void {
            int produced = 0;
            for (int j = 0; j < count; ++j) {
                test.enqueue(j);
                ++produced;
            }
            
            // Give time to consumers to finish their jobs.
            // On GCC 4.5 strangely destroys performance by 2s or more
            // I wasn't expecting that much
            this_thread::sleep_for(chrono::seconds(1));
        } );
    }
    
    // Consumers
    for (int i = 0; i < nThreads; ++i) {
        consumers[i] = thread([&,i] () -> void {
            int res, read = 0;
            map<int, int> m;

            while (!over) {
                while (!test.dequeue(res))
                    if (over) break;
               if (over) break;     
                m[res%100]++;
                ++read;
            }

            results[i] = m;
            totals[i] = read;
        });
    }
    
    for (int i = 0; i < nThreads; ++i)
        producers[i].join();

    cout << "Production finished." << endl;
    over = true;

    // Results
    for (int i = 0; i < nThreads; ++i) {
        consumers[i].join();
        cout << "Consumer " << i << " read " << totals[i] << endl;
    }
    
    for (int i = 0; i < 100; ++i) {
        int tot = 0;
        for (int j = 0; j < nThreads; ++j)
            tot += results[j][i];

        if (tot != count * nThreads / 100)
            cout << "  !! " << i << ": " << tot << " -- " << count * nThreads / 100 << endl;
    }
}


void SimpleTestTBB()
{
    cout << "TBB Simple Test" << endl;
    int count = 100 * 1000 * 1000;
    
    concurrent_queue<int> test;
    
    int produced = 0;
    int consumed = 0;
    
    thread producer([&] () -> void {
        int tmp = 0;

        for (int i = 0; i < count; ++i) {
            test.push(i);
            ++tmp;
        }
        produced = tmp;
    } );
    
    thread consumer([&] () -> void {
        int res = 0, last = -1, tmp = 0;

        for (int i = 0; i < count; ++i) {
            while (!test.try_pop(res)) { }
            ++tmp;
            assert(res == last + 1);
            last = res;
        }
        consumed = tmp;
    } );
    
    producer.join();
    consumer.join();
    
    cout << "Produced : " << produced << endl;
    cout << "Consumed : " << consumed << endl;
}



// Simple test with memory allocation allowed
void SimpleTest()
{
    cout << "boost::lockfree Simple Test" << endl;
    int count = 100 * 1000 * 1000;
    
    fifo<int> test;
    
    int produced = 0;
    int consumed = 0;
    
    thread producer([&] () -> void {
        int tmp = 0;

        for (int i = 0; i < count; ++i) {
            test.enqueue(i);
            ++tmp;
        }
        produced = tmp;
    } );
    
    thread consumer([&] () -> void {
        int res = 0, last = -1, tmp = 0;

        for (int i = 0; i < count; ++i) {
            while (!test.dequeue(res)) { }
            ++tmp;
            assert(res == last + 1);
            last = res;
        }
        consumed = tmp;
    } );
    
    producer.join();
    consumer.join();
    
    cout << "Produced : " << produced << endl;
    cout << "Consumed : " << consumed << endl;
}